/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 > File:            S.powaix/src/syscomp/sysdefs.p
 > Purpose:
 > Author:          John Gibson, Jan 15 1998
 */

/* -------------------------------------------------------------------------

				DEFINITIONS FOR MACHINE/OPERATING SYSTEM

--------------------------------------------------------------------------*/

section;

global constant macro (

	POPC_SYSDEFS_LOADED = true,


;;; --- SYSTEM NAME -----------------------------------------------------

	MACHINE			= [[power]],


;;; --- MACHINE = POWER[PC] -----------------------------------------------

	PROCESSOR		= [[power]],	;;; type of cpu
	POWER			= true,

	;;; Values for machine and C data types are defined in mcdata.p,
	;;; and can be overidden here if necessary

	BIG_ENDIAN		= true,
	STACK_ALIGN_BITS= 64,			;;; alignment in bits for stack frames


;;; === OPERATING SYSTEM (UNIX BSD) ==================================


	UNIX = true,
	BERKELEY = 4.3,
	LINUX = 2.0,
	LINUX_ELF = true,
	POSIX1 = 199309,			;;; probably later than this ...
	OPERATING_SYSTEM = [[unix linux ^LINUX elf posix {^POSIX1}]],

	SHARED_LIBRARIES = true,

	VPAGE_OFFS = 4096,		;;; word address offset of a virtual page

	;;; LOWEST_ADDRESS:
	LOWEST_ADDRESS = 0,

	;;; UNIX_USRSTACK:
	;;; 	????

	UNIX_USRSTACK  = 16:FC000000,

	;;; Procedures to get and set the memory break and return the REAL end of
	;;; memory. (We always need the real end to ensure that the end of the
	;;; user stack is always at the true end of memory, so that user stack
	;;; underflow produces a memory access violation.)

	GET_REAL_BREAK =
		[procedure(); _extern sbrk(_0)@(b.r->vpage) endprocedure],

	SET_REAL_BREAK =
		[procedure(_break) -> _break;
			lvars _break = _break@(w.r->vpage);
			if _extern brk(_break@(w->b)) == _-1 then
				_-1 -> _break
			endif
		endprocedure],


;;; --- OTHER ------------------------------------------------------------


	;;; C returns "float" as double
	C_FLOAT_RESULT_SINGLE = false,

	;;; list of procedures to be optimised as subroutine calls
	;;; format of entries is
	;;;		[<pdr name> <nargs> <nresults> <subroutine name>]

	SUBROUTINE_OPTIMISE_LIST =
		[[
			[prolog_newvar			0 1 _prolog_newvar]
			[datakey				1 1 _datakey]
			[prolog_deref			1 1 _prolog_deref]
			[conspair				2 1 _conspair]
		]],

	;;; if GEN_SVB_CONSTANTS is true, constants in _special_var_block will be
	;;; generated by Poplink, and recognised in the production of M-code
	;;; operands by Popc (providing genproc.p defines SPECIAL_VAR_BLOCK_REG).

	GEN_SVB_CONSTANTS	= true,

);

define :inline CACHEFLUSH(_ptr, _nbytes);
	if $-Sys$- _cache_flush_needed then
		$-Sys$- _cache_flush(_ptr, _nbytes)
	endif
enddefine;

endsection;		/* $- */
